/*
 *	adsbox - ADS-B decoder software
 *	(c) atty 2011 romulmsg@mail.ru
 *
 */

#include <stdio.h>
#include <stdlib.h> 
#include <fcntl.h>
#include <termios.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <pthread.h>
#include <time.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include "misc.h"

#define BUFLEN 256
#define PORT 30003
#define BAUDRATE B115200
#define DEVICE "/dev/ttyS0"

pthread_mutex_t mux;
pthread_cond_t cond_var;
char gdata[BUFLEN];
int dmode = 0;
float homelat = 0;
float homelon = 0;
_Bool record = 0;
_Bool play = 0;
float playrate = 1;
FILE * rd;
int delay = 0;

double sec_between(struct timespec now, struct timespec then) {
	double result;
	result = now.tv_sec - then.tv_sec + 
		(double)(now.tv_nsec - then.tv_nsec) / 1000000000;
	return (result);
}

void *decoder(void *arg) { /* decoder thread */
	int fd = (int) arg;
	unsigned char rawbuf[BUFLEN];
	unsigned char msg[128];
	struct tm *now;
	struct timespec ts;

	while (1) {
		bzero(rawbuf, BUFLEN);
		read(fd, rawbuf, BUFLEN);
		if (strlen(rawbuf) == 17) {
			if (record) {
				clock_gettime(CLOCK_REALTIME, &ts);
				now = localtime(&ts.tv_sec);
				fprintf(rd, "%02d:%02d:%02d.%03d %02d.%02d.%04d %s", \
					now->tm_hour, now->tm_min, now->tm_sec, ts.tv_nsec/1000000, now->tm_mday, now->tm_mon+1, now->tm_year+1900, rawbuf);
				fflush(rd);
			}
			if (decode_message56(rawbuf, msg)) {
				pthread_mutex_lock(&mux);
				bzero(gdata, BUFLEN);
				memcpy(gdata, msg, strlen(msg));
				pthread_cond_broadcast(&cond_var);
				pthread_mutex_unlock(&mux);
				bzero(msg, 128);
			}
		}

		if (strlen(rawbuf) == 31) {
			if (record) {
				clock_gettime(CLOCK_REALTIME, &ts);
				now = localtime(&ts.tv_sec);
				fprintf(rd, "%02d:%02d:%02d.%03d %02d.%02d.%04d %s", \
					now->tm_hour, now->tm_min, now->tm_sec, ts.tv_nsec/1000000, now->tm_mday, now->tm_mon+1, now->tm_year+1900, rawbuf);
				fflush(rd);
			}
			if (decode_message(rawbuf, msg)) {
				pthread_mutex_lock(&mux);
				bzero(gdata, BUFLEN);
				memcpy(gdata, msg, strlen(msg));
				pthread_cond_broadcast(&cond_var);
				pthread_mutex_unlock(&mux);
				bzero(msg, 128);
			}
		}
	}
	pthread_exit(0);
}

void *player(void *arg) { /* player-decoder thread */
	unsigned char rawbuf[BUFLEN];
	unsigned char msg[128];
	time_t tnow, tprev;
	struct tm tm;
	int r;
	struct timespec ts_now, ts_prev;
	struct timespec fts_now, fts_prev;
	
	for(r=0;r<delay;r++) sleep(1);

	bzero(rawbuf, BUFLEN);
	r = fscanf(rd, "%u:%u:%u.%u %u.%u.%u %s\n", 
		&tm.tm_hour,&tm.tm_min,&tm.tm_sec,&fts_prev.tv_nsec,
		&tm.tm_mday,&tm.tm_mon,&tm.tm_year, 
		&rawbuf);
	tm.tm_mon = tm.tm_mon - 1;
	tm.tm_year = tm.tm_year - 1900;
	fts_prev.tv_sec = mktime(&tm);
	fts_prev.tv_nsec = fts_prev.tv_nsec * 1000000;

	while (r != EOF) {
		strcat(rawbuf, "\n");
		if (strlen(rawbuf) == 17) {
			if (decode_message56(rawbuf, msg)) {
				pthread_mutex_lock(&mux);
				bzero(gdata, BUFLEN);
				memcpy(gdata, msg, strlen(msg));
				pthread_cond_broadcast(&cond_var);
				pthread_mutex_unlock(&mux);
				bzero(msg, 128);
			}
		}

		if (strlen(rawbuf) == 31) {
			if (decode_message(rawbuf, msg)) {
				pthread_mutex_lock(&mux);
				bzero(gdata, BUFLEN);
				memcpy(gdata, msg, strlen(msg));
				pthread_cond_broadcast(&cond_var);
				pthread_mutex_unlock(&mux);
				bzero(msg, 128);
			}
		}
		bzero(rawbuf, BUFLEN);
		
		r = fscanf(rd, "%u:%u:%u.%u %u.%u.%u %s\n", 
			&tm.tm_hour,&tm.tm_min,&tm.tm_sec,&fts_prev.tv_nsec,
			&tm.tm_mday,&tm.tm_mon,&tm.tm_year,
			&rawbuf);
		tm.tm_mon = tm.tm_mon - 1;
		tm.tm_year = tm.tm_year - 1900;
		fts_now.tv_sec = mktime(&tm);
		fts_now.tv_nsec = fts_now.tv_nsec * 1000000;

		tprev = time(NULL);
		clock_gettime(CLOCK_REALTIME, &ts_prev);
		while (sec_between(ts_now, ts_prev) * playrate < sec_between(fts_now, fts_prev)) {
			usleep(100000);
			clock_gettime(CLOCK_REALTIME, &ts_now);
		}
		fts_prev = fts_now;	
	}
	DEBUG("EOF Ctrl-C to exit.\n");
	pthread_exit(0);
}

void *servlet(void *arg) /* process client connection thread */
{
	char buf[2];
	char cdata[BUFLEN];
	struct timespec to;

	int s = (int) arg;

	while (recv(s, buf, 1, MSG_DONTWAIT)) {
		pthread_mutex_lock(&mux);
		to.tv_sec = time(NULL) + 1;
		to.tv_nsec = 0;
		pthread_cond_timedwait(&cond_var, &mux, &to);
		if (strncmp(cdata, gdata, strlen(gdata))) {
			bzero(cdata, BUFLEN);
			memcpy(cdata, gdata, strlen(gdata));
			pthread_mutex_unlock(&mux);
			send(s, cdata, strlen(cdata), 0);
		}
		pthread_mutex_unlock(&mux);
	}
	close(s);
	pthread_exit(0);
}

int main(int ac, char *av[]) {
	extern char * optarg;
	char s[16] = DEVICE;
	struct sockaddr_in addr;
	int fd, sd, sd1, en = 1;
	struct termios oldtio, newtio;
	int pid, sid;
	char fname[64];
	int c;

	pthread_mutex_init(&mux, NULL);
	bzero(fname, 64);

	printf("adsbox ADS-B decoder version %s.\n", VERSION);
	
	while ((c = getopt_long(ac, av, "ds:l:g:r:p:x:z:")) != -1) {
		switch (c) {
		case 'd':
			dmode = 1;
			break;
		case 's':
			strncpy(s, optarg, 16);
			break;
		case 'l':
			homelat = atof(optarg);
			break;
		case 'g':
			homelon = atof(optarg);
			break;
		case 'r':
			record = 1;
			play = !record;
			strncpy(fname, optarg, 64);
			break;
		case 'p':
			play = 1;
			record = !play;
			strncpy(fname, optarg, 64);
			break;
		case 'x':
			playrate = atof(optarg);
			break;
		case 'z':
			delay = atoi(optarg);
			break;
		case '?':
			printf("%s [options]\n-s input serial device\n-d be a daemon\n-l our latitude\n-g our longitude\n-r [file] record ADS-B data to file\n-p [file] play recordered file\n-x [n] play n times faster\n-z [n] snooze for n seconds before play", av[0]);
			exit(0);
		default:
			break;
		}
	}
	if ((homelat != 0) && (homelon != 0))
		printf("Your lat %2.5f lon %3.5f\n", homelat, homelon);
	
	if (record) {
		if((rd = fopen(fname, "w+")) == NULL) {
			printf("Unable to open %s\n", fname);
			exit(1);
		}
		printf("Recording data to %s.\n", fname);
	}

	if (play) {
		if((rd = fopen(fname, "r")) == NULL) {
			printf("Unable to open %s\n", fname);
			exit(1);
		}
		printf("Playing data from %s file.\n", fname);
	}

	/* COMPORT STUFF
	 */
	if (!play) {
		fd = open(s, O_RDONLY | O_NOCTTY);
		if (fd < 0)
			panic("open device");

		printf("Using %s as input serial device.\n", s);

		tcgetattr(fd, &oldtio); // TODO restore old termio

		bzero(&newtio, sizeof(newtio));
		newtio.c_cflag &= ~PARENB;
			newtio.c_cflag &= ~CSTOPB;
		newtio.c_cflag &= ~CSIZE;
			newtio.c_cflag |= BAUDRATE | CS8 | CLOCAL | CREAD;
		newtio.c_iflag = IGNPAR | ICRNL; // ignore parity error
		newtio.c_oflag = 0;

		newtio.c_lflag = ICANON;

		newtio.c_cc[VTIME] = 0;
			newtio.c_cc[VMIN] = 1;
		newtio.c_cc[VEOL] = 0;
		newtio.c_cc[VEOL2] = 0;

		tcflush(fd, TCIFLUSH);
		tcsetattr(fd, TCSANOW, &newtio);
	}
	/* COMPORT STUFF
	 */

	//char msg[256];
	//decode_message("8D4000FB5875444148D9A70B023A", msg);
	//exit(0);

	/*--- create socket ---*/
	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (sd < 0)
		panic("socket");
	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &en, sizeof(en)) != 0)
		panic("setsockopt");

	/*--- bind  ---*/
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(PORT);
	addr.sin_addr.s_addr = htonl(INADDR_ANY); /* any interface */
	if (bind(sd, (struct sockaddr*) &addr, sizeof(addr)) != 0)
		panic("bind");

	if (dmode) { /* daemonize */
		pid = fork();
		if (pid < 0)
			panic("fork()");

		if (pid > 0)
			exit(0);

		umask(0);

		sid = setsid();

		if (sid < 0)
			panic("setsid()");

		chdir("/");

		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
	}

	/*--- make into listener with 10 slots ---*/
	if (listen(sd, 10) != 0)
		panic("listen()")

	/*--- begin waiting for connections ---*/
	else {
		pthread_t child;

		if (play)
			pthread_create(&child, NULL, player, NULL); /* start player thread */
		else
			pthread_create(&child, NULL, decoder, (void *) fd); /* start decoder thread */

		while (1) /* process incoming clients */
		{
			sd1 = accept(sd, 0, 0); /* accept connection */
			pthread_create(&child, NULL, servlet, (void *) sd1); /* start connection thread */
		}
	}
	exit(0);
}
